import numpy as np
import scipy.optimize
import ipopt

# Main assumptions: log mean TD (Chen approximation), heat loss

x_vars = ['alpha3', 'T', 'Tstar1', 'Tstar2', 'Tstar3', 'The1', 'The2', 'The3', 'Q1', 'Q2', 'Q3', 'Qloss1', 'Qloss2',
          'Qloss3', 'T1', 'T2', 'T3']
u_vars = ['alpha1', 'alpha2']

meas_sets = {
    1: ['T0', 'T1', 'T2', 'T3', 'The1', 'The2', 'The3'],
    2: ['T0', 'w0', 'wh1', 'wh2', 'wh3', 'Th1', 'Th2', 'Th3', 'UA1', 'UA2', 'UA3', 'alpha1', 'alpha2']
}

Ti_max = 1500
grad_abseps = 1e-8;
h0 = 1e-4; hrate = 10.0;

def model_f(x, u, par):
    alpha1 = u[0]
    alpha2 = u[1]

    alpha3 = x[0]
    T = x[1]
    Tstar1 = x[2]
    Tstar2 = x[3]
    Tstar3 = x[4]
    The1 = x[5]
    The2 = x[6]
    The3 = x[7]
    Q1 = x[8]
    Q2 = x[9]
    Q3 = x[10]
    Qloss1 = x[11]
    Qloss2 = x[12]
    Qloss3 = x[13]
    T1 = x[14]
    T2 = x[15]
    T3 = x[16]

    T0 = par['T0'];
    w0 = par['w0'];
    Th1 = par['Th1'];
    Th2 = par['Th2'];
    Th3 = par['Th3'];
    wh1 = par['wh1'];
    wh2 = par['wh2'];
    wh3 = par['wh3'];
    UA1 = par['UA1'];
    UA2 = par['UA2'];
    UA3 = par['UA3'];
    Ts = par['Ts'];
    h1 = par['h1'];
    h2 = par['h2'];
    h3 = par['h3'];

    Tprod1 = ((Th1 - Tstar1) * (The1 - T0) * ((Th1 - Tstar1) + (The1 - T0)) / 2)
    Tprod2 = ((Th2 - Tstar2) * (The2 - T0) * ((Th2 - Tstar2) + (The2 - T0)) / 2)
    Tprod3 = ((Th3 - Tstar3) * (The3 - T0) * ((Th3 - Tstar3) + (The3 - T0)) / 2)

    dTlm1 = np.sign(Tprod1) * np.abs(Tprod1) ** (1 / 3);
    dTlm2 = np.sign(Tprod2) * np.abs(Tprod2) ** (1 / 3);
    dTlm3 = np.sign(Tprod3) * np.abs(Tprod3) ** (1 / 3);

    f0 = - T + alpha1 * T1 + alpha2 * T2 + alpha3 * T3;
    f01 = alpha1 + alpha2 + alpha3 - 1;
    f11 = - Q1 + w0 * alpha1 * (Tstar1 - T0);
    f12 = - Q2 + w0 * alpha2 * (Tstar2 - T0);
    f13 = - Q3 + w0 * alpha3 * (Tstar3 - T0);
    f21 = - Q1 + UA1 * dTlm1;
    f22 = - Q2 + UA2 * dTlm2;
    f23 = - Q3 + UA3 * dTlm3;
    f31 = - Q1 + wh1 * (Th1 - The1);
    f32 = - Q2 + wh2 * (Th2 - The2);
    f33 = - Q3 + wh3 * (Th3 - The3);
    f41 = - Qloss1 + w0 * alpha1 * (T1 - Tstar1);
    f42 = - Qloss2 + w0 * alpha2 * (T2 - Tstar2);
    f43 = - Qloss3 + w0 * alpha3 * (T3 - Tstar3);
    f51 = - Qloss1 + h1 * (Ts - T1);
    f52 = - Qloss2 + h2 * (Ts - T2);
    f53 = - Qloss3 + h3 * (Ts - T3);

    f = np.array([f0, f01, f11, f12, f13, f21, f22, f23, f31, f32, f33, f41, f42, f43, f51, f52, f53]);
    return f

def model_J(x, u, par):
    T = x[1:2]

    J = -T
    return J

def output(u, par):
    nx = len(x_vars)
    nu = len(u_vars)

    x0 = np.array( [0.33] + [(par['T0']+par['Th1'])/2, (par['T0']+par['Th2'])/2, (par['T0']+par['Th3'])/2]*2 + [par['T0']] * (nx - 2*(nu+1) - 1) )

    lbx = np.array([0] + [None] * (nx - 1));  # constraint on inputs and first state (last flow split)
    ubx = np.array([None] + [None] * (nx - 1));  # upper limit on splits is not necessary, but will automatically be satisfied

    lbx[(2):(2 + 3 * 2)] = par['T0'];  # constraint on temperatures
    ubx[(2):(2 + 3)] = np.array([par['Th1'], par['Th2'], par['Th3']]);  # constraint on temperatures
    ubx[(2 + 3):(2 + 3 * 2)] = np.array([par['Th1'], par['Th2'], par['Th3']]);  # constraint on temperatures

    bds = [(lbxi, ubxi) for lbxi, ubxi in zip(lbx, ubx)]
    f_x = lambda x: model_f(x, u, par)
    nlcon = {'type': 'eq', 'fun': f_x}
    J_x = lambda x: model_J(x, u, par)
    # res = ipopt.minimize_ipopt(J_x, x0, jac=None, bounds=bds, constraints=nlcon )#, tol=None)
    res = scipy.optimize.minimize(J_x, x0, method='SLSQP', bounds=bds, constraints=nlcon)  # , tol=None)

    return {'x': res.x, 'success': res.success}

def cost(u, par):
    xdict = output(u, par)
    # assert xdict['success']
    J = model_J(xdict['x'], u, par)
    return {'J': J, 'success': xdict['success']}

def grad(u, par):
    Ju = np.zeros(u.shape)
    for i in range(len(u)):
        h = h0
        hvec = np.zeros(u.shape)
        hvec[i] = h
        Juiest = (cost(u + hvec / 2, par)['J'] - cost(u - hvec / 2, par)['J']) / h
        while True:
            h = h / hrate
            hvec[i] = h
            Juiestnew = (cost(u + hvec / 2, par)['J'] - cost(u - hvec / 2, par)['J']) / h
            if np.abs(Juiestnew - Juiest) < grad_abseps:
                Ju[i] = Juiestnew
                break
            Juiest = Juiestnew
    return {'grad': Ju, 'success': True};  # should do something if derivative fails?

def output_meas(meas_set, u, par):
    meas_vars = meas_sets[meas_set];
    y = np.zeros((len(meas_vars),));
    out = output(u, par);
    x = out['x'];
    for i, var in enumerate(meas_vars):
        if var in par:
            y[i] = par[var];
        elif var in u_vars:
            y[i] = u[u_vars.index(var)];
        elif var in x_vars:
            y[i] = x[x_vars.index(var)];
        else:
            y[i] = np.nan;
    return {'y': y, 'success': out['success']}

def optim(par):
    nx = len(x_vars)
    nu = len(u_vars)

    lbx = np.array([0] * (nu + 1) + [None] * (nx - 1));  # constraint on inputs and first state (last flow split)
    ubx = np.array([None] * (nu + 1) + [None] * (
                nx - 1));  # upper limit on splits is not necessary, but will automatically be satisfied

    lbx[(nu + 2):(nu + 2 + 3 * 2)] = par['T0'];  # constraint on temperatures
    ubx[(nu + 2):(nu + 2 + 3)] = np.array([min(Ti_max, t) for t in [par['Th1'], par['Th2'], par['Th3']] ]);  # constraint on temperatures
    ubx[(nu + 2 + 3):(nu + 2 + 3 * 2)] = np.array([par['Th1'], par['Th2'], par['Th3']]);  # constraint on temperatures

    bds = [(lbxi, ubxi) for lbxi, ubxi in zip(lbx, ubx)]

    # x0 = np.array([0.33] * (nu + 1) + [par['T0'] + 5] * (nx - 1))
    x0 = np.array( [0.33]*(nu+1) + [(par['T0']+par['Th1'])/2, (par['T0']+par['Th2'])/2, (par['T0']+par['Th3'])/2]*2 + [par['T0']] * (nx - 2*(nu+1) - 1) )

    f_ux = lambda ux: model_f(ux[nu:], ux[:nu], par)
    nlcon = {'type': 'eq', 'fun': f_ux}
    J_ux = lambda ux: model_J(ux[nu:], ux[:nu], par)
    # res = ipopt.minimize_ipopt(J_ux, x0, jac=None, bounds=bds, constraints=nlcon )#, tol=None)
    res = scipy.optimize.minimize(J_ux, x0, method='SLSQP', bounds=bds, constraints=nlcon)  # , tol=None)
    return {'u': res.x[:nu], 'x': res.x[-nx:], 'success': res.success}
